subroutine hwarpinv2(t,ikn,mkn,rng,winv,errflag,m,p)
!Inverse Hermite-spline warping function
!
!INPUT:
!   T       (m)     Time grid where inverse is computed
!   IKN     (p)     Individual knots (or pseudo-landmarks), excluding endpoints
!   MKN     (p)     Mean knots, excluding endpoints
!   RNG     (2)     Endpoints of time interval
!
!OUTPUT:
!   WINV    (m)     Inverse warping function
!	ERRFLAG (integer)	0 if there were no errors, 1 otherwise

implicit none
! declare arguments
integer, intent(in) :: m, p
double precision, intent(in) :: t(m), ikn(p), mkn(p), rng(2)
double precision, intent(out) :: winv(m)
integer, intent(out) :: errflag
! declare locals
integer, parameter :: Ngrid = 500
double precision :: tt(Ngrid),phi(p+2,Ngrid),psi(p+2,Ngrid),alpha(p+2),beta(p+2),w(Ngrid), a, b
integer :: i, j, idx(Ngrid)

winv = t
errflag = 0
a = rng(1)
b = rng(2)
if (any(ikn<=a).or.any(ikn>=b).or.any(mkn<=a).or.any(mkn>=b)) then
	print "(A)", " Knots outside range in HWARPINV"
    errflag = 1
    return
else if (any(ikn(2:p)<ikn(1:p-1))) then
	print "(A)", " Individual knot sequence not monotone increasing in HWARPINV"
    errflag = 1
    return
else if (any(mkn(2:p)<mkn(1:p-1))) then
	print "(A)", " Mean knot sequence not monotone increasing in HWARPINV"
    errflag = 1
    return
else
	!Fine-grid warping function
	forall(i=1:Ngrid) tt(i) = a + (b-a)*dble(i-1)/dble(Ngrid-1)
	call basis(tt,mkn,a,b,phi,psi,Ngrid,p)
	alpha(1) = a
	alpha(2:p+1) = ikn
	alpha(p+2) = b
	call FCtangents(mkn,alpha,a,b,beta,p)
	w = matmul(alpha,phi) + matmul(beta,psi)
	!Inverse warping function
	do i = 1,m
		if (all(w<t(i))) then
			winv(i) = b
		else
			j = minloc(w,1,w>=t(i))		! j = find(w>=t(i),1,'first') in Matlab
	    	winv(i) = tt(j)
    	end if
   	end do
end if

end subroutine hwarpinv2


!!! ----------- Auxiliary functions ----------- !!!

subroutine basis(t,theta,a,b,phi,psi,m,p)
!Hermite basis functions
!(Note: THETA are internal knots -- exclude endpoints)
!	t		(m)
!	theta	(p)
!	a, b	(scalars)
!	phi		(p+2 x m)
!	psi		(p+2 x m)

implicit none
!declare arguments
integer, intent(in) :: m, p
double precision, intent(in) :: t(m), theta(p), a, b
double precision, intent(out) :: phi(p+2,m), psi(p+2,m)
!declare locals
integer :: k
double precision :: theta2(p+2), ones(m), aux1(m), aux2(m)

if (any(theta>b).or.any(theta<a)) then
	print "(A)", " Knots outside range in HWARPINV/BASIS"
else
	! Basis functions
	phi = 0.d0
	psi = 0.d0
	theta2(1) = a
	theta2(2:p+1) = theta
	theta2(p+2) = b
	call h00((t-theta2(1))/(theta2(2)-theta2(1)),m,phi(1,:))
	call h10((t-theta2(1))/(theta2(2)-theta2(1)),m,aux1)
	psi(1,:) = aux1*(theta2(2)-theta2(1))
	do k = 2,p+1
		ones = 0.d0
		where(t==theta2(k)) ones = 1.d0
       	call h00((t-theta2(k))/(theta2(k+1)-theta2(k)),m,aux1)
       	call h00((theta2(k)-t)/(theta2(k)-theta2(k-1)),m,aux2)
       	phi(k,:) = aux1 + aux2 - ones
       	call h10((t-theta2(k))/(theta2(k+1)-theta2(k)),m,aux1)
       	call h10((theta2(k)-t)/(theta2(k)-theta2(k-1)),m,aux2)
    	psi(k,:) = aux1*(theta2(k+1)-theta2(k)) - aux2*(theta2(k)-theta2(k-1))
	end do
	call h00((theta2(p+2)-t)/(theta2(p+2)-theta2(p+1)),m,phi(p+2,:))
	call h10((theta2(p+2)-t)/(theta2(p+2)-theta2(p+1)),m,aux1)
	psi(p+2,:) = -aux1*(theta2(p+2)-theta2(p+1))
end if

end subroutine basis

! -----

subroutine h00(t,m,ans)
implicit none
integer, intent(in) :: m
double precision, intent(in) :: t(m)
double precision, intent(out) :: ans(m)
ans = 0.d0
where((t>=0.d0).and.(t<=1.d0)) ans = (1.d0+2.d0*t)*(1.d0-t)**2
end subroutine h00

! -----

subroutine h10(t,m,ans)
implicit none
integer, intent(in) :: m
double precision, intent(in) :: t(m)
double precision, intent(out) :: ans(m)
ans = 0.d0
where((t>=0.d0).and.(t<=1.d0)) ans = t*(1.d0-t)**2
end subroutine h10

! -----

subroutine FCtangents(tau,alpha,a,b,beta,p)
!Fritsch-Carlson tangets for Hermite interpolant
!	tau		(p)
!	alpha	(p+2)
!	a, b	(scalars)
!	beta	(p+2)

implicit none
! declare arguments
integer, intent(in) :: p
double precision, intent(in) :: tau(p), alpha(p+2), a, b
double precision, intent(out) :: beta(p+2)
! declare locals
integer :: k
double precision :: xx(p+2), Delta(p+1)
double precision :: ak, bk, tauk

xx(1) = a
xx(2:p+1) = tau
xx(p+2) = b
beta = 0.d0
Delta = (alpha(2:p+2)-alpha(1:p+1))/(xx(2:p+2)-xx(1:p+1))
beta(1) = Delta(1)
beta(2:p+1) = 0.5d0*Delta(2:p+1) + 0.5d0*Delta(1:p)
beta(p+2) = Delta(p+1)
do k = 1,p+1
    if (Delta(k)==0.d0) then
        beta(k) = 0.d0
        beta(k+1) = 0.d0
    else
        ak = beta(k)/Delta(k)
        bk = beta(k+1)/Delta(k)
        tauk = 1.d0
        if (ak**2+bk**2>9.d0) tauk = 3.d0/sqrt(ak**2+bk**2)
        beta(k) = tauk*ak*Delta(k)
        beta(k+1) = tauk*bk*Delta(k)
	end if
end do
end subroutine FCtangents


