function [modpar,f,pred,errflag] = WFRe(data,modpar,baspar,N,mxit,dsp,sub0)
%function [modpar,f,pred,errflag] = WFRe(data,modpar,baspar,N,mxit,dsp,sub0)
% Warped Functional Regression (equal time grids; x and y inputs in matrix form)
% Fits model Z = muZ + A*(W-muW) + E, where W=[U;theta_x] and Z=[V;theta_y].
%
% INPUT
%   data: Data (raw measurements). Struct with the following fields:
%       s: Time grid of functional covariate (1 x nux)
%       x: Functional covariate observations (n x nux)
%       t: Time grid of functional response (1 x nuy)
%       y: Functional response observations (n x nuy)
%   modpar: Initial model parameters. Struct with the following fields:
%       mx: Basis coefficients for mean function of the covariate process (qx x 1)
%       my: Basis coefficients for mean function of the response process (qy x 1)
%       Cx: Basis coefficients for K-L components of the covariate process (qx x px)
%       Cy: Basis coefficients for K-L components of the response process (qy x py)
%       s2x: Variance of x-measurement errors (scalar)
%       s2y: Variance of y-measurement errors (scalar)
%       A: Regression matrix (dy x dx)
%       Se: Covariance matrix of error term (dy x dy)
%       taux0: Mean x-warping knots (1 x rx)
%       tauy0: Mean y-warping knots (1 x ry)
%       Sw: Joint covariance matrix of W = [U;theta_x] (dx x dx)
%   baspar: Basis parameters. Struct with the following fields:
%       ors: Spline order (4 is cubic) for x-space basis
%       nks: Number of knots for x-space basis
%       rs: Range of S used for estimation (2 x 1 vector with lower and upper bounds)
%       ort: Spline order (4 is cubic) for y-space basis
%       nkt: Number of knots for y-space basis
%       rt: Range of T used for estimation (2 x 1 vector with lower and upper bounds)
%   N:      Number of Monte Carlo samples for integration (integer)
%   mxit:   Maximum number of iterations (integer)
%   dsp:    Display updates on each iteration ('on'/'off' character variable)
%   sub0:   Optional argument, (ns0 x 2)-matrix of subindices (i,j) indicating
%            elements of A that are constrained to be zero (useful for testing)
%
%OUTPUT
%   modpar: Same as input, with ML estimators of the parameters
%   f: Likelihood value for each individual (n x 1)
%   pred: Individual predictors. Struct with the following fields:
%       U: K-L scores of explanatory process (n x px)
%       theta_x: x-warping parameters (Jupp-transformed knots) (n x rx)
%       tau_x: x-warping knots (n x rx)
%       V: K-L scores of response process (n x py)
%       theta_y: y-warping parameters (Jupp-transformed knots) (n x ry)
%       tau_y: y-warping knots (n x ry)
%   errflag: 0/1 error flag; if null likelihood values are encountered during
%       iterations, errflag = 1 and program exits returning empty f and pred
%
%External programs called: COMPSUMS (calls Fortran routine), BSPL, FIND_C, JUPP
%
% Version: January 2012

% Input check
if nargin==7
    ind0 = sub2ind(size(modpar.A),sub0(:,1),sub0(:,2));
else
    ind0 = [];
end
n = size(data.x,1);
nux = sum((data.s>=baspar.rs(1))&(data.s<=baspar.rs(2)));
nuy = sum((data.t>=baspar.rt(1))&(data.t<=baspar.rt(2)));

% Initialization of variables
grN = 500;
sgrid = linspace(baspar.rs(1),baspar.rs(2),grN+1);
sknots = linspace(baspar.rs(1),baspar.rs(2),baspar.nks+2);
B = bspl(sgrid,baspar.ors,sknots,0);
Jx = B'*B*(sgrid(2)-sgrid(1));
Jx1 = inv(sqrtm(Jx));

tgrid = linspace(baspar.rt(1),baspar.rt(2),grN+1);
tknots = linspace(baspar.rt(1),baspar.rt(2),baspar.nkt+2);
B = bspl(tgrid,baspar.ort,tknots,0);
Jy = B'*B*(tgrid(2)-tgrid(1));
Jy1 = inv(sqrtm(Jy));

[qx,px] = size(modpar.Cx);
[qy,py] = size(modpar.Cy);
[dy,dx] = size(modpar.A);

% MC basic samples
Wmc0 = normrnd(0,1,[dx N]);
Emc0 = normrnd(0,1,[dy N]);

% Initial log-likelihood
errflag = 0;
modpar.A(ind0) = 0;
s = compsums(data,modpar,baspar,Wmc0,Emc0);
if ~all(isfinite(log(s.f)))
    errflag = 1;
    disp('Some density values are zero -- WFR will return to invoking function')
    f = [];
    pred = [];
    return
end
L = mean(log(s.f));
if strcmp(dsp,'on')
    disp(['Initial Loglik = ' num2str(L)])
end
% Iterations
err = 1;
iter = 0;
while err>1e-3 && iter<mxit
    iter = iter+1;
    L0 = L;
    % Estimator updates
    switch mod(iter,2)
        case 1
            % Structural parameters
            modpar.Sw = s.Sw0w0/n;
            modpar.Sw(1:px,1:px) = diag(diag(s.Sw0w0(1:px,1:px)/n));
            modpar.Se = diag(s.Sr1_2)/n;
            modpar.A = s.Sz0w0/s.Sw0w0;
            if py>1
                sqSw = sqrtm(modpar.Sw);
                M2 = sqSw\(s.Sz0w0(1:py,:)'/n);
                D12 = modpar.Se(1:py,1:py);
                for i = 1:3
                    [U,dummy] = svd(M2*diag(1./diag(modpar.Se(1:py,1:py)))*D12,0);
                    D12 = diag(diag(U'*M2));
                end
                modpar.A(1:py,:) = (sqSw\U*D12)';
            end
            modpar.A(ind0) = 0;
        case 0
            % Nuisance parameters
            modpar.mx = s.SBBx\s.SBRx;
            modpar.my = s.SBBy\s.SBRy;
            modpar.Cx(:,1) = Jx1*find_c(Jx1*s.SuuBBx(:,:,1)*Jx1,Jx1*s.SuBxr6(:,1));
            for j = 2:px
                modpar.Cx(:,j) = Jx1*find_c(Jx1*s.SuuBBx(:,:,j)*Jx1,Jx1*s.SuBxr6(:,j),Jx1*modpar.Cx(:,1:j-1));
            end            
            modpar.Cy(:,1) = Jy1*find_c(Jy1*s.SvvBBy(:,:,1)*Jy1,Jy1*s.SvByr7(:,1));
            for j = 2:py
                modpar.Cy(:,j) = Jy1*find_c(Jy1*s.SvvBBy(:,:,j)*Jy1,Jy1*s.SvByr7(:,j),Jy1*modpar.Cy(:,1:j-1));
            end
    end
    modpar.s2x = s.Snormx2/(n*nux);
    modpar.s2y = s.Snormy2/(n*nuy);
    % Update sums
    s = compsums(data,modpar,baspar,Wmc0,Emc0);
    % Convergence check
    L = mean(log(s.f));
    if ~all(isfinite(log(s.f)))
        errflag = 1;
        disp('Some density values are zero -- WFR will return to invoking function')
        f = [];
        pred = [];
        return
    end
    err = abs(L-L0)/abs(L0);
    if strcmp(dsp,'on')
        disp(['Iter: ' num2str(iter) ', Loglik = ' num2str(L) ', Error = ' num2str(err)])
    end    
end

% Output

f = s.f;
tau_x = jupp(s.pred_w(:,px+1:dx),baspar.rs(1),baspar.rs(2),'inverse');
tau_y = jupp(s.pred_z(:,py+1:dy),baspar.rt(1),baspar.rt(2),'inverse');
pred = struct('U',s.pred_w(:,1:px),'theta_x',s.pred_w(:,px+1:dx),...
    'V',s.pred_z(:,1:py),'theta_y',s.pred_z(:,py+1:dy),'tau_x',tau_x,'tau_y',tau_y);
