function [modpar,f,pred,errflag] = WFRe0(data,baspar,taux0,tauy0,N,mxit,dsp)
%function [modpar,f,pred,errflag] = WFRe0(data,baspar,taux0,tauy0,N,mxit,dsp)
% Initial Warped Functional Regression (equal time grids; x and y inputs in matrix form)
% Fits model Z = muZ + A*(W-muW) + E, where W = theta_x and Z = theta_y
% (This is the initial mean-and-warping-only model, used mainly to initialize WFRe)
%
% INPUT
%   data: Data (raw measurements). Struct with the following fields:
%       s: Time grid of functional covariate (1 x nux)
%       x: Functional covariate observations (n x nux)
%       t: Time grid of functional response (1 x nuy)
%       y: Functional response observations (n x nuy)
%   baspar: Basis parameters. Struct with the following fields:
%       ors: Spline order (4 is cubic) for x-space basis
%       nks: Number of knots for x-space basis
%       rs: Range of S used for estimation (2 x 1 vector with lower and upper bounds)
%       ort: Spline order (4 is cubic) for y-space basis
%       nkt: Number of knots for y-space basis
%       rt: Range of T used for estimation (2 x 1 vector with lower and upper bounds)
%   taux0:  Mean x-warping knots (1 x rx)
%   tauy0:  Mean y-warping knots (1 x ry)
%   N:      Number of Monte Carlo samples for integration (integer)
%   mxit:   Maximum number of iterations (integer)
%   dsp:    Display updates on each iteration ('on'/'off' character variable)
%
%OUTPUT
%   modpar: ML estimators of the parameters. Struct with the following fields:
%       mx: Basis coefficients for mean function of the covariate process (qx x 1)
%       my: Basis coefficients for mean function of the response process (qy x 1)
%       s2x: Variance of x-measurement errors (scalar)
%       s2y: Variance of y-measurement errors (scalar)
%       A: Regression matrix (ry x rx)
%       Se: Covariance matrix of error term (ry x ry)
%       Sw: Covariance matrix of W = theta_x (rx x rx)
%       taux0:  Same as input (redundant but necessary for sub-functions)
%       tauy0:  Same as input (redundant but necessary for sub-functions)
%   f: Likelihood value for each individual (n x 1)
%   pred: Individual predictors. Struct with the following fields:
%       theta_x: x-warping parameters (Jupp-transformed knots) (n x rx)
%       tau_x: x-warping knots (n x rx)
%       theta_y: y-warping parameters (Jupp-transformed knots) (n x ry)
%       tau_y: y-warping knots (n x ry)
%   errflag: 0/1 error flag; if null likelihood values are encountered during
%       iterations, errflag = 1 and program exits returning empty f and pred
%
%External programs called: COMPSUMS0 (calls Fortran routine), BSPL, JUPP
%
% Version: January 2012

% Initialization of variables
n = size(data.x,1);
rx = length(taux0);
ry = length(tauy0);

knots = linspace(baspar.rs(1),baspar.rs(2),baspar.nks+2);
inrange = (data.s>=baspar.rs(1))&(data.s<=baspar.rs(2));
B = bspl(data.s(inrange),baspar.ors,knots,0);
nux = size(B,1);
mx = (B'*B)\(B'*mean(data.x(:,inrange))');
s2x = mean(mean((data.x(:,inrange)-ones(n,1)*(B*mx)').^2));

knots = linspace(baspar.rt(1),baspar.rt(2),baspar.nkt+2);
inrange = (data.t>=baspar.rt(1))&(data.t<=baspar.rt(2));
B = bspl(data.t(inrange),baspar.ort,knots,0);
nuy = size(B,1);
my = (B'*B)\(B'*mean(data.y(:,inrange))');
s2y = mean(mean((data.y(:,inrange)-ones(n,1)*(B*my)').^2));

A = zeros(ry,rx);
Sw = eye(rx);
Se = eye(ry);

modpar = struct('mx',mx,'my',my,'s2x',s2x,'s2y',s2y,'A',A,'Se',Se,'Sw',Sw,...
    'taux0',taux0,'tauy0',tauy0);

% MC basic samples
dx = rx;
dy = ry;
Wmc0 = normrnd(0,1,[dx N]);
Emc0 = normrnd(0,1,[dy N]);

% Initial log-likelihood
errflag = 0;
s = compsums0(data,modpar,baspar,Wmc0,Emc0);
if ~all(isfinite(log(s.f)))
    errflag = 1;
    disp('Some density values are zero -- WFR will return to invoking function')
    f = [];
    pred = [];
    return
end
L = mean(log(s.f));
if strcmp(dsp,'on')
    disp(['Initial Loglik = ' num2str(L)])
end

% Iterations
err = 1;
iter = 0;
while err>1e-3 && iter<mxit
    iter = iter+1;
    L0 = L;
    % Estimator updates
    switch mod(iter,2)
        case 1
            % Structural parameters
            modpar.Sw = s.Sw0w0/n;
            modpar.Se = diag(s.Sr1_2)/n;
            modpar.A = s.Sz0w0/s.Sw0w0;
        case 0
            % Nuisance parameters
            modpar.mx = s.SBBx\s.SBRx;
            modpar.my = s.SBBy\s.SBRy;
    end
    modpar.s2x = s.Snormx2/(n*nux);
    modpar.s2y = s.Snormy2/(n*nuy);
    % Update sums
    s = compsums0(data,modpar,baspar,Wmc0,Emc0);
    % Convergence check
    L = mean(log(s.f));
    if ~all(isfinite(log(s.f)))
        errflag = 1;
        disp('Some density values are zero -- WFR will return to invoking function')
        f = [];
        pred = [];
        return
    end
    err = abs(L-L0)/abs(L0);
    if strcmp(dsp,'on')
        disp(['Iter: ' num2str(iter) ', Loglik = ' num2str(L) ', Error = ' num2str(err)])
    end    
end

% Output
f = s.f;
tau_x = jupp(s.pred_w,baspar.rs(1),baspar.rs(2),'inverse');
tau_y = jupp(s.pred_z,baspar.rt(1),baspar.rt(2),'inverse');
pred = struct('theta_x',s.pred_w,'theta_y',s.pred_z,'tau_x',tau_x,'tau_y',tau_y);
