function [stA,AV] = asymp(data,modpar,baspar,N)
%function [stA,AV] = asymp(data,modpar,baspar,N)
% Joint asymptotic covariance matrix of [vec(A'),v(Sw)]
% INPUT: 
%   As in WFRe.m
% OUTPUT:
%   stA: Standard deviations of elements of A in corresponding matrix form
%        (already divided by sqrt(n))
%   AV:  Joint asymp. cov. matrix of [vec(A'),v(Sw)] (not divided by n)
% External functions used: DUPLICATION, JUPP, HWARPINV, BSPL, COMPV_GTW (Fortran)
% Version: May 2012

n = size(data.x,1);
[qy,py] = size(modpar.Cy);
[dy,dx] = size(modpar.A);

V = comp_V(data,modpar,baspar,N);

if py<2
    AV = inv(V);
else
    m = (py-1)*py/2;
    p = dx*dy+dx*(dx+1)/2;
    B = zeros(m,p);
    Dup = duplication(dx);
    k = 0;
    for i = 2:py
        ei = zeros(1,py);
        ei(i) = 1;
        for j = 1:i-1
            k = k+1;
            ej = zeros(1,py);
            ej(j) = 1;
            B(k,1:py*dx) = modpar.A(i,:)*modpar.Sw*kron(ej,eye(dx)) + ...
                modpar.A(j,:)*modpar.Sw*kron(ei,eye(dx));
            B(k,dx*dy+1:end) = kron(modpar.A(j,:),modpar.A(i,:))*Dup;
        end
    end
    PB = eye(p) - B'*((B*B')\B);
    [G,dummy] = svd(PB);
    G2 = G(:,1:(p-m));
    AV = G2*((G2'*V*G2)\G2');
end

stA = sqrt(reshape(diag(AV(1:dx*dy,1:dx*dy)),[dx dy])'/n);


% --------------------------------------
%   Auxiliary Function
% --------------------------------------


function V = comp_V(data,modpar,baspar,N)
% Computes "V", the inverse of Fisher's Information matrix, used for the
%   computation of the WFR asymptotic covariance matrix
% INPUT: 
%   As in WFRe.m
% OUTPUT:
%   V
n = size(data.x,1);
[qx,px] = size(modpar.Cx);
[qy,py] = size(modpar.Cy);
[dy,dx] = size(modpar.A);
Wmc0 = normrnd(0,1,[dx N]);
Emc0 = normrnd(0,1,[dy N]);
muthx = jupp(modpar.taux0,baspar.rs(1),baspar.rs(2),'direct')';
muthy = jupp(modpar.tauy0,baspar.rt(1),baspar.rt(2),'direct')';
muw = [zeros(px,1);muthx];
muz = [zeros(py,1);muthy];
Wmc = muw*ones(1,N) + real(sqrtm(modpar.Sw))*Wmc0;
Zmc = muz*ones(1,N) + modpar.A*(Wmc-muw*ones(1,N)) + sqrtm(modpar.Se)*Emc0;
[w0w0,z0w0] = compv_gtw(data.s,data.t,data.x,data.y,baspar.rs,baspar.rt, ...
    ones(1,baspar.nks),ones(1,baspar.nkt),ones(1,baspar.ors),ones(1,baspar.ort), ...
    Wmc,Zmc,modpar.mx,modpar.my,modpar.Cx,modpar.Cy,modpar.s2x,modpar.s2y,...
    modpar.taux0,modpar.tauy0);
Se_1 = inv(modpar.Se);
Sw_1 = inv(modpar.Sw);
Dup = duplication(dx);
mU = zeros(dx*dy+dx*(dx+1)/2,1);
V = zeros(dx*dy+dx*(dx+1)/2);
for i = 1:n
    U1 = reshape(z0w0(:,:,i)'*Se_1-w0w0(:,:,i)*modpar.A'*Se_1,[dx*dy 1]);
    U2 = -.5*Dup'*reshape(Sw_1-Sw_1*w0w0(:,:,i)*Sw_1,[dx^2 1]);
    mU = mU + [U1;U2]/n;
    V = V + [U1;U2]*[U1' U2']/n;
end
V = V - mU*mU';
clear compv_gtw

