function [Aboot,errflag] = bootWFRe(data,modpar,baspar,N,mxit,dsp,NB,type)
%function [Aboot,errflag] = bootWFRe(data,modpar,baspar,N,mxit,dsp,NB,type)
% Bootstrap estimators of A (Warped Functional Regression coefficient)
% Input and model specifications are as in WFRe.m, in addition to:
%   NB:     Number of bootstrap samples
%   type:   'wild' or 'param', for wild or parametric bootstrap, respectively.
% Output is a (dy x dx x NB) array of A estimates plus error flag.
%
%External programs called: COMPSUMS (calls Fortran routine), BSPL, FIND_C, JUPP, HWARPINV
%
% Version: May 2012

% Input check
n = size(data.x,1);
nux = sum((data.s>=baspar.rs(1))&(data.s<=baspar.rs(2)));
nuy = sum((data.t>=baspar.rt(1))&(data.t<=baspar.rt(2)));

% Initialization of variables
grN = 500;
sgrid = linspace(baspar.rs(1),baspar.rs(2),grN+1);
sknots = linspace(baspar.rs(1),baspar.rs(2),baspar.nks+2);
B = bspl(sgrid,baspar.ors,sknots,0);
Jx = B'*B*(sgrid(2)-sgrid(1));
Jx1 = inv(sqrtm(Jx));

tgrid = linspace(baspar.rt(1),baspar.rt(2),grN+1);
tknots = linspace(baspar.rt(1),baspar.rt(2),baspar.nkt+2);
B = bspl(tgrid,baspar.ort,tknots,0);
Jy = B'*B*(tgrid(2)-tgrid(1));
Jy1 = inv(sqrtm(Jy));

[qx,px] = size(modpar.Cx);
[qy,py] = size(modpar.Cy);
[dy,dx] = size(modpar.A);

x0 = data.x;
y0 = data.y;
modpar0 = modpar;
Aboot = zeros(dy,dx,NB);
errflag = 0;

if ~(strcmp(type,'wild')||strcmp(type,'param'))
    errflag = 1;
    disp('Variable TYPE must be either "wild" or "param"')
    return
end

% MC basic samples
Wmc0 = normrnd(0,1,[dx N]);
Emc0 = normrnd(0,1,[dy N]);

% Bootstrap replications
for ib = 1:NB
    if strcmp(dsp,'on')
        disp(' ')
        disp(['Bootstrap sample ' num2str(ib)])
        disp(' ')
    end
    if strcmp(type,'wild')
        ii = randi(n,[n 1]);
        data.x = x0(ii,:);
        data.y = y0(ii,:);
    else
        [x_boot,y_boot] = genbootdata(data,modpar0,baspar);
        data.x = x_boot;
        data.y = y_boot;
    end
    modpar = modpar0;
    % Initial log-likelihood
    s = compsums(data,modpar,baspar,Wmc0,Emc0);
    if ~all(isfinite(log(s.f)))
        errflag = 1;
        disp('Some density values are zero')
        return
    end
    L = mean(log(s.f));
    if strcmp(dsp,'on')
        disp(['Initial Loglik = ' num2str(L)])
    end
    % Iterations
    err = 1;
    iter = 0;
    while err>1e-3 && iter<mxit
        iter = iter+1;
        L0 = L;
        % Estimator updates
        switch mod(iter,2)
            case 1
            % Structural parameters
            modpar.Sw = s.Sw0w0/n;
            modpar.Sw(1:px,1:px) = diag(diag(s.Sw0w0(1:px,1:px)/n));
            modpar.Se = diag(s.Sr1_2)/n;
            modpar.A = s.Sz0w0/s.Sw0w0;
            if py>1
                sqSw = sqrtm(modpar.Sw);
                M2 = sqSw\(s.Sz0w0(1:py,:)'/n);
                D12 = modpar.Se(1:py,1:py);
                for i = 1:3
                    [U,dummy] = svd(M2*diag(1./diag(modpar.Se(1:py,1:py)))*D12,0);
                    D12 = diag(diag(U'*M2));
                end
                modpar.A(1:py,:) = (sqSw\U*D12)';
            end
            case 0
            % Nuisance parameters
            modpar.mx = s.SBBx\s.SBRx;
            modpar.my = s.SBBy\s.SBRy;
            modpar.Cx(:,1) = Jx1*find_c(Jx1*s.SuuBBx(:,:,1)*Jx1,Jx1*s.SuBxr6(:,1));
            for j = 2:px
                modpar.Cx(:,j) = Jx1*find_c(Jx1*s.SuuBBx(:,:,j)*Jx1,Jx1*s.SuBxr6(:,j),Jx1*modpar.Cx(:,1:j-1));
            end            
            modpar.Cy(:,1) = Jy1*find_c(Jy1*s.SvvBBy(:,:,1)*Jy1,Jy1*s.SvByr7(:,1));
            for j = 2:py
                modpar.Cy(:,j) = Jy1*find_c(Jy1*s.SvvBBy(:,:,j)*Jy1,Jy1*s.SvByr7(:,j),Jy1*modpar.Cy(:,1:j-1));
            end
        end
        modpar.s2x = s.Snormx2/(n*nux);
        modpar.s2y = s.Snormy2/(n*nuy);
        % Update sums
        s = compsums(data,modpar,baspar,Wmc0,Emc0);
        % Convergence check
        L = mean(log(s.f));
        if ~all(isfinite(log(s.f)))
            errflag = 1;
            disp('Some density values are zero')
            return
        end
        err = abs(L-L0)/abs(L0);
        if strcmp(dsp,'on')
            disp(['Iter: ' num2str(iter) ', Loglik = ' num2str(L) ', Error = ' num2str(err)])
        end    
    end
    Aboot(:,:,ib) = modpar.A;
end

% ------------------------
% Auxiliary Function
% ------------------------

function [x_boot,y_boot] = genbootdata(data,modpar,baspar)

px = size(modpar.Cx,2);
py = size(modpar.Cy,2);
muthx = jupp(modpar.taux0,baspar.rs(1),baspar.rs(2),'direct');
muw = [zeros(px,1);muthx'];
muthy = jupp(modpar.tauy0,baspar.rt(1),baspar.rt(2),'direct');
muz = [zeros(py,1);muthy'];

[dy,dx] = size(modpar.A);
[n,m] = size(data.x);
w_boot = repmat(muw,[1 n]) + sqrtm(modpar.Sw)*randn(dx,n);
e_boot = sqrtm(modpar.Se)*randn(dy,n);
z_boot = repmat(muz,[1 n]) + modpar.A*(w_boot-repmat(muw,[1 n])) + e_boot;

x_boot = zeros(n,m);
y_boot = zeros(n,m);
taux = jupp(w_boot(px+1:dx,:)',baspar.rs(1),baspar.rs(2),'inverse');
tauy = jupp(z_boot(py+1:dy,:)',baspar.rt(1),baspar.rt(2),'inverse');
knt_s = linspace(baspar.rs(1),baspar.rs(2),baspar.nks+2);
knt_t = linspace(baspar.rt(1),baspar.rt(2),baspar.nkt+2);
for i = 1:n
    s_star = hwarpinv(data.s,taux(i,:),modpar.taux0,baspar.rs);
    t_star = hwarpinv(data.t,tauy(i,:),modpar.tauy0,baspar.rt);
    Bx = bspl(s_star,baspar.ors,knt_s,0);
    By = bspl(t_star,baspar.ort,knt_t,0);
    x_boot(i,:) = (Bx*modpar.mx + Bx*modpar.Cx*w_boot(1:px,i) + sqrt(modpar.s2x)*randn(m,1))';
    y_boot(i,:) = (By*modpar.my + By*modpar.Cy*z_boot(1:py,i) + sqrt(modpar.s2y)*randn(m,1))';
end






