function winv = hwarpinv(t,ikn,mkn,rng)
%function winv = hwarpinv(t,ikn,mkn,rng)
%Inverse Hermite-spline warping function
%
%INPUT:
%   T       (1 x m)     Time grid where inverse is computed
%   IKN     (1 x p)     Individual knots (or pseudo-landmarks), excluding endpoints
%   MKN     (1 x p)     Mean knots, excluding endpoints
%   RNG     (1 x 2)     Endpoints of time interval
%
%OUTPUT:
%   WINV    (1 x m)     Inverse warping function

% Initialization
m = length(t);
if size(t,1)>1
    t = t';
end
winv = NaN(1,m);
a = rng(1);
b = rng(2);
if size(ikn,1)>1
    ikn = ikn';
end
if size(mkn,1)>1
    mkn = mkn';
end
if any(ikn<=a)||any(ikn>=b)||any(mkn<=a)||any(mkn>=b)
    disp('Knots outside range')
    return
end

% Fine-grid warping function
Ngrid = max(10*m,300);
tt = linspace(a,b,Ngrid);
[phi,psi] = basis(tt,mkn,a,b);
alpha = [a ikn b];
beta = FCtangents(mkn,alpha,a,b);
w = alpha*phi + beta*psi;

% Inverse warping function
for i = 1:m
    j = find(w>=t(i),1,'first');
    winv(i) = tt(j);
end

%%% ----------- Auxiliary functions ----------- %%%

function [phi,psi] = basis(t,theta,a,b)
%function [phi,psi] = basis(t,theta,a,b)
%Hermite basis functions
%(Note: THETA are internal knots -- exclude endpoints)

% Initialization
p = length(theta);
if size(theta,1)>1
    theta = theta';
end
m = length(t);
if size(t,1)>1
    t = t';
end
if any(theta>b|theta<a)
    error('Some knots are outside the range')
end

% Hermite polynomials
h00 = @(t)((1+2*t).*(1-t).^2.*(t>=0).*(t<=1));
h10 = @(t)(t.*(1-t).^2.*(t>=0).*(t<=1));

% Basis functions
phi = zeros(p+2,m);
psi = zeros(p+2,m);
theta = [a theta b];  % Endpoints added as extreme knots
phi(1,:) = h00((t-theta(1))/(theta(2)-theta(1)));
psi(1,:) = h10((t-theta(1))/(theta(2)-theta(1)))*(theta(2)-theta(1));
for k = 2:p+1
    phi(k,:) = h00((t-theta(k))/(theta(k+1)-theta(k))) + ...
        h00((theta(k)-t)/(theta(k)-theta(k-1))) - (t==theta(k));
    psi(k,:) = h10((t-theta(k))/(theta(k+1)-theta(k)))*(theta(k+1)-theta(k)) - ...
        h10((theta(k)-t)/(theta(k)-theta(k-1)))*(theta(k)-theta(k-1));
end
phi(p+2,:) = h00((theta(p+2)-t)/(theta(p+2)-theta(p+1)));
psi(p+2,:) = -h10((theta(p+2)-t)/(theta(p+2)-theta(p+1)))*(theta(p+2)-theta(p+1));

function beta = FCtangents(tau,alpha,a,b)
%Fritsch-Carlson tangets for Hermite interpolant
p = length(tau);
xx = [a tau b];
n = size(alpha,1);
beta = zeros(n,p+2);
Delta = (alpha(:,2:p+2)-alpha(:,1:p+1))./(ones(n,1)*(xx(2:p+2)-xx(1:p+1)));
beta(:,1) = Delta(:,1);
beta(:,2:p+1) = 0.5*Delta(:,2:p+1) + 0.5*Delta(:,1:p);
beta(:,p+2) = Delta(:,p+1);
for k = 1:p+1
    %if Delta(k)==0
        I0 = (Delta(:,k)==0);
        beta(I0,k) = 0;
        beta(I0,k+1) = 0;
    %else
        ak = beta(~I0,k)./Delta(~I0,k);
        bk = beta(~I0,k+1)./Delta(~I0,k);
        tauk = ones(size(ak));
            %if outside monotonicity region
            I2 = (ak.^2+bk.^2>9);
            tauk(I2) = 3./sqrt(ak(I2).^2+bk(I2).^2);
        %rescaled betas
        beta(~I0,k) = tauk.*ak.*Delta(~I0,k);
        beta(~I0,k+1) = tauk.*bk.*Delta(~I0,k);
end
