function [y,v,tau_y] = pred_WFRe(s,x,t,modpar,baspar,N)
%function [y,v,tau_y] = pred_WFRe(s,x,t,modpar,baspar,N)
% Predict Y from X using Warped Functional Regression model (equal time grid)
% Fits model Z = muZ + A*(W-muW) + E, where W=[U;theta_x] and Z=[V;theta_y].
%
% INPUT
%   s: Time grid of functional covariate (1 x nux)
%   x: Functional covariate (n x nux)
%   t: Time grid of functional response (1 x nuy)
%   modpar: Estimated model parameters (struct, same fields as in WFRe)
%   baspar: Basis parameters (struct, same fields as in WFRe)
%   N:      Number of Monte Carlo samples for integration (integer)
%
%OUTPUT
%   y: Predicted functional response (n x nuy)
%   v: Predicted K-L scores of the response process (n x py)
%   tau_y: Predicted y-warping knots (n x ry)
%
% External programs called: JUPP, BSPL
%
% Version: July 2014

% Input check
if any((t<baspar.rt(1))|(t>baspar.rt(2)))
    error('Values of T must be within baspar.rt(1) and baspar.rt(2)')
end
% Constants
n = size(x,1);
[qx,px] = size(modpar.Cx);
[qy,py] = size(modpar.Cy);
dy = size(modpar.A,1);
% Computation of W and Z
[w,eflag] = find_w(s,x,modpar,baspar,N);    % w is n x dx
if eflag
    y = [];
    v = [];
    tau_y = [];
    return
end
muthx = jupp(modpar.taux0,baspar.rs(1),baspar.rs(2),'direct')';
muthy = jupp(modpar.tauy0,baspar.rt(1),baspar.rt(2),'direct')';
muw = [zeros(px,1);muthx];
v = (modpar.A(1:py,:)*(w'-muw*ones(1,n)))';     % n x py
thetay = (muthy*ones(1,n) + modpar.A(py+1:dy,:)*(w'-muw*ones(1,n)))';   % n x ry
tau_y = jupp(thetay,baspar.rt(1),baspar.rt(2),'inverse');       % n x ry
% Computation of Y
tknots = linspace(baspar.rt(1),baspar.rt(2),baspar.nkt+2);
y = NaN(n,length(t));
for i = 1:n
    t_star = hwarpinv(t,tau_y(i,:),modpar.tauy0,baspar.rt);
    if ~any(isnan(t_star))
        By = bspl(t_star,baspar.ort,tknots,0);
        y(i,:) = (By*modpar.my + By*modpar.Cy*v(i,:)')';
    end
end

% -------- AUXILIARY FUNCTIONS ---------

function [w,eflag] = find_w(s,x,modpar,baspar,N)
% Initialization
eflag = 0;
n = size(x,1);
[qx,px] = size(modpar.Cx);
[dy,dx] = size(modpar.A);
% MC samples
Wmc0 = normrnd(0,1,[dx N]);
muthx = jupp(modpar.taux0,baspar.rs(1),baspar.rs(2),'direct')';
muw = [zeros(px,1);muthx];
Wmc = muw*ones(1,N) + sqrtm(modpar.Sw)*Wmc0;
tauxmc = jupp(Wmc(px+1:dx,:)',baspar.rs(1),baspar.rs(2),'inverse')';
% MC integrals
jx = (s>=baspar.rs(1))&(s<=baspar.rs(2));
nux = sum(jx);
sknots = linspace(baspar.rs(1),baspar.rs(2),baspar.nks+2);
f = zeros(n,1);
w = zeros(n,dx);
for j = 1:N
    s_star = hwarpinv(s(jx),tauxmc(:,j)',modpar.taux0,baspar.rs);
    if ~any(isnan(s_star))
        Bx = bspl(s_star,baspar.ors,sknots,0);
        mux = Bx*(modpar.mx+modpar.Cx*Wmc(1:px,j));
        fxcond = (1/(2*pi*modpar.s2x)^(nux/2))* ...
            exp(-0.5*sum((x(:,jx)-ones(n,1)*mux').^2,2)/modpar.s2x);
        f = f + fxcond/N;
        w = w + fxcond*Wmc(:,j)'/N;
    else
        eflag = 1;
        return
    end
end
w = w./(f*ones(1,dx));

function winv = hwarpinv(t,ikn,mkn,rng)
%function winv = hwarpinv(t,ikn,mkn,rng)
%Inverse Hermite-spline warping function
%
%INPUT:
%   T       (1 x m)     Time grid where inverse is computed
%   IKN     (1 x p)     Individual knots (or pseudo-landmarks), excluding endpoints
%   MKN     (1 x p)     Mean knots, excluding endpoints
%   RNG     (1 x 2)     Endpoints of time interval
%
%OUTPUT:
%   WINV    (1 x m)     Inverse warping function

% Initialization
m = length(t);
if size(t,1)>1
    t = t';
end
winv = NaN(1,m);
a = rng(1);
b = rng(2);
if size(ikn,1)>1
    ikn = ikn';
end
if size(mkn,1)>1
    mkn = mkn';
end
if any(ikn<=a)||any(ikn>=b)||any(mkn<=a)||any(mkn>=b)
    disp('Knots outside range')
    return
end

% Fine-grid warping function
Ngrid = 500;
tt = linspace(a,b,Ngrid);
[phi,psi] = basis(tt,mkn,a,b);
alpha = [a ikn b];
beta = FCtangents(mkn,alpha,a,b);
w = alpha*phi + beta*psi;

% Inverse warping function
for i = 1:m
    j = find(w>=t(i),1,'first');
    winv(i) = tt(j);
end

function [phi,psi] = basis(t,theta,a,b)
%function [phi,psi] = basis(t,theta,a,b)
%Hermite basis functions
%(Note: THETA are internal knots -- exclude endpoints)

% Initialization
p = length(theta);
if size(theta,1)>1
    theta = theta';
end
m = length(t);
if size(t,1)>1
    t = t';
end
if any(theta>b|theta<a)
    error('Some knots are outside the range')
end

% Hermite polynomials
h00 = @(t)((1+2*t).*(1-t).^2.*(t>=0).*(t<=1));
h10 = @(t)(t.*(1-t).^2.*(t>=0).*(t<=1));

% Basis functions
phi = zeros(p+2,m);
psi = zeros(p+2,m);
theta = [a theta b];  % Endpoints added as extreme knots
phi(1,:) = h00((t-theta(1))/(theta(2)-theta(1)));
psi(1,:) = h10((t-theta(1))/(theta(2)-theta(1)))*(theta(2)-theta(1));
for k = 2:p+1
    phi(k,:) = h00((t-theta(k))/(theta(k+1)-theta(k))) + ...
        h00((theta(k)-t)/(theta(k)-theta(k-1))) - (t==theta(k));
    psi(k,:) = h10((t-theta(k))/(theta(k+1)-theta(k)))*(theta(k+1)-theta(k)) - ...
        h10((theta(k)-t)/(theta(k)-theta(k-1)))*(theta(k)-theta(k-1));
end
phi(p+2,:) = h00((theta(p+2)-t)/(theta(p+2)-theta(p+1)));
psi(p+2,:) = -h10((theta(p+2)-t)/(theta(p+2)-theta(p+1)))*(theta(p+2)-theta(p+1));

function beta = FCtangents(tau,alpha,a,b)
%Fritsch-Carlson tangets for Hermite interpolant
p = length(tau);
xx = [a tau b];
n = size(alpha,1);
beta = zeros(n,p+2);
Delta = (alpha(:,2:p+2)-alpha(:,1:p+1))./(ones(n,1)*(xx(2:p+2)-xx(1:p+1)));
beta(:,1) = Delta(:,1);
beta(:,2:p+1) = 0.5*Delta(:,2:p+1) + 0.5*Delta(:,1:p);
beta(:,p+2) = Delta(:,p+1);
for k = 1:p+1
    %if Delta(k)==0
        I0 = (Delta(:,k)==0);
        beta(I0,k) = 0;
        beta(I0,k+1) = 0;
    %else
        ak = beta(~I0,k)./Delta(~I0,k);
        bk = beta(~I0,k+1)./Delta(~I0,k);
        tauk = ones(size(ak));
            %if outside monotonicity region
            I2 = (ak.^2+bk.^2>9);
            tauk(I2) = 3./sqrt(ak(I2).^2+bk(I2).^2);
        %rescaled betas
        beta(~I0,k) = tauk.*ak.*Delta(~I0,k);
        beta(~I0,k+1) = tauk.*bk.*Delta(~I0,k);
end

