function [estim,f,pred,errflag] = runWFRe(s,x,t,y,a,b,taux0,tauy0,px,py,nknots,mxit,N,dsp)
%[estim,f,pred,errflag] = runWFRe(s,x,t,y,a,b,taux0,tauy0,px,py,nknots,mxit,N,dsp)
% Runs WFRe sequentially, starting with a mean-and-warping-only model and
%   building up to a (px,py)-component model
% After the initial model, the sequence is run in the following order: 
%   (1,1),(2,1),...,(px,1),(px,2),...,(px,py)
%
% INPUT
%   s:      Time grid for X     (1 x nux)
%   x:      Covariate curves    (n x nux)
%   t:      Time grid for Y     (1 x nuy)
%   y:      Response curves     (n x nuy)
%   a:      Scalar ([a,b] is time range of interest)
%   b:      Scalar ([a,b] is time range of interest)
%   taux0:  X-warping knots (1 x rx)
%   tauy0:  Y-warping knots (1 x ry)
%   px:     Number of X-pc's of final model (integer; can be 0)
%   py:     Number of Y-pc's of final model (integer; can be 0)
%   nknots: Number of spline knots for means and pcs (integer)
%           (Cubic splines with equispaced knots will be used)
%   mxit:   Maximum number of iterations per model (integer)
%   N:      Number of Monte Carlo samples for computation of integrals (integer)
%   dsp:    Display iteration updates (string: 'yes' or 'no')
%
% OUTPUT
%   estim: Cell struct with parameter estimators. Cell "k" corresponds to the k-th fitted
%       model in the order given above. Typically you only need the last cell.
%       The variables in the struct are the same as in "modpar" returned by WFRe.m,
%       except for taux0 and tauy0 which are not included.
%   f: As returned by WFRe.m for the *last* model.
%   pred: As returned by WFRe.m for the *last* model.
%   errflag: 0/1 variable; 1 if there was an error and execution was interrupted.
%
% External programs called: WFRe0, WFRe, 
%
% Version: March 2012
%
% Output initialization
nmodls = 1+px+(py-1);
mx = cell(1,nmodls);
my = cell(1,nmodls);
Cx = cell(1,nmodls);
Cy = cell(1,nmodls);
s2x = cell(1,nmodls);
s2y = cell(1,nmodls);
A = cell(1,nmodls);
Se = cell(1,nmodls);
Sw = cell(1,nmodls);
estim = [];
f = [];
pred = [];
errflag = 1;

% Input check
[nr,nc] = size(s);
if  nr>1 && nc>1
    disp('S must be a vector (common time grid for all covariates is assumed)')
    return
elseif nr>1
    s = s';
end
nux = length(s);
[nr,nc] = size(x);
if nc~=nux
    disp('# of columns of X must equal length of S')
    return
else
    n = nr;
end
[nr,nc] = size(t);
if  nr>1 && nc>1
    disp('T must be a vector (common time grid for all responses is assumed)')
    return
elseif nr>1
    t = t';
end
nuy = length(t);
[nr,nc] = size(y);
if nc~=nuy
    disp('# of columns of Y must equal length of T')
    return
elseif nr~=n
    disp('# of rows of X and Y must be equal')
    return
end
if a>=b
    disp('A must be less than B')
    return
end
if all(s<a) || all(s>b)
    disp('Range [a,b] does not include any values of S')
    return
end
if all(t<a) || all(t>b)
    disp('Range [a,b] does not include any values of T')
    return
end
[nr,nc] = size(taux0);
if  nr>1 && nc>1
    disp('TAUX0 must be a vector')
    return
elseif nr>1
    taux0 = taux0';
end
rx = length(taux0);
if taux0(1)<=a || taux0(rx)>=b || any(taux0(1:rx-1)>=taux0(2:rx))
    disp('TAUX0 must be a strictly increasing sequence in [a,b]')
    return
end
[nr,nc] = size(tauy0);
if  nr>1 && nc>1
    disp('TAUY0 must be a vector')
    return
elseif nr>1
    tauy0 = tauy0';
end
ry = length(tauy0);
if tauy0(1)<=a || tauy0(ry)>=b || any(tauy0(1:ry-1)>=tauy0(2:ry))
    disp('TAUY0 must be a strictly increasing sequence in [a,b]')
    return
end
if (px>0&&py==0)||(px==0&&py>0)
    disp('PX and PY must be both 0 or both >= 1')
    return
end

% Input structs for WFRe0 and WFRe
data = struct('s',s,'x',x,'t',t,'y',y);
ord = 4;
baspar = struct('ors',ord,'nks',nknots,'rs',[a b],'ort',ord,'nkt',nknots,'rt',[a b]);

% Run initial (0,0)-pc model
if strcmp(dsp,'on')
    disp(' ')
    disp('Running initial model ... ')
    disp(' ')
end
[modpar,f,pred,errflag] = WFRe0(data,baspar,taux0,tauy0,N,mxit,dsp);
if errflag==1
    return
end
mx{1} = modpar.mx;
my{1} = modpar.my;
s2x{1} = modpar.s2x;
s2y{1} = modpar.s2y;
A{1} = modpar.A;
Se{1} = modpar.Se;
Sw{1} = modpar.Sw;

% Run (i,1)-pc models
q = ord + nknots;
for i = 1:px
    if i==1
        modpar.Cx = mx{1};
        modpar.Cy = my{1};
        modpar.A = [0,zeros(1,rx);zeros(ry,1),A{1}];
        modpar.Sw = [s2x{1},zeros(1,rx);zeros(rx,1),Sw{1}];
        modpar.Se = diag([s2y{1};diag(Se{1})]);
    else        
        modpar.Cx = [Cx{1+i-1},ones(q,1)];
        modpar.A = [A{1+i-1}(:,1:i-1),zeros(ry+1,1),A{1+i-1}(:,i:end)];
        modpar.Sw = [Sw{1+i-1}(1:i-1,1:i-1),zeros(i-1,1),Sw{1+i-1}(1:i-1,i:end); ...
                     zeros(1,i-1),s2x{1+i-1},zeros(1,rx); ...
                     Sw{1+i-1}(i:end,1:i-1),zeros(rx,1),Sw{1+i-1}(i:end,i:end)];
    end
    if strcmp(dsp,'on')
        disp(' ')
        disp(['Running model with (' num2str(i) ',1) components ... '])
        disp(' ')
    end
    [modpar,f,pred,errflag] = WFRe(data,modpar,baspar,N,mxit,dsp);
    if errflag==1
        return
    end
    mx{1+i} = modpar.mx;
    my{1+i} = modpar.my;
    Cx{1+i} = modpar.Cx;
    Cy{1+i} = modpar.Cy;
    s2x{1+i} = modpar.s2x;
    s2y{1+i} = modpar.s2y;
    A{1+i} = modpar.A;
    Se{1+i} = modpar.Se;
    Sw{1+i} = modpar.Sw;
end

% Run (px,j)-pc models
for j = 2:py
    modpar.Cy = [Cy{1+px+j-2},ones(q,1)];
    modpar.A = [A{1+px+j-2}(1:j-1,:);zeros(1,px+rx);A{1+px+j-2}(j:end,:)];
    modpar.Se = diag([diag(Se{1+px+j-2}(1:j-1,1:j-1));s2y{1+px+j-2};diag(Se{1+px+j-2}(j:end,j:end))]);
    if strcmp(dsp,'on')
        disp(' ')
        disp(['Running model with (' num2str(px) ',' num2str(j) ') components ... '])
        disp(' ')
    end
    [modpar,f,pred,errflag] = WFRe(data,modpar,baspar,N,mxit,dsp);
    if errflag==1
        return
    end
    mx{1+px+j-1} = modpar.mx;
    my{1+px+j-1} = modpar.my;
    Cx{1+px+j-1} = modpar.Cx;
    Cy{1+px+j-1} = modpar.Cy;
    s2x{1+px+j-1} = modpar.s2x;
    s2y{1+px+j-1} = modpar.s2y;
    A{1+px+j-1} = modpar.A;
    Se{1+px+j-1} = modpar.Se;
    Sw{1+px+j-1} = modpar.Sw;
end

% Output
estim = struct('mx',mx,'my',my,'Cx',Cx,'Cy',Cy,'s2x',s2x,'s2y',s2y,'A',A,'Se',Se,'Sw',Sw);
